import random
import uuid
from pathlib import Path

import pandas as pd
import pytest

from framelink.core import FramelinkModel, FramelinkPipeline, F, T

TEST_ROOT_DIR = Path(__file__).parent
DATA_DIR = TEST_ROOT_DIR.parent.parent / "data"

random.seed(0)


@pytest.fixture
def src_frame() -> F[T]:
    def src_frame(_: FramelinkPipeline) -> pd.DataFrame:
        """
        Mock data frame for testing purposes
        :return: mock data
        """
        n = 10
        return pd.DataFrame(
            {
                "id_col": list(range(n)),
                "string_col": [random.choice(["apple", "banana", "cherry"]) for _ in range(n)],
                "uuid_col": [str(uuid.uuid4()) for _ in range(n)],
                "int_col": [random.randint(0, 100) for _ in range(n)],
            }
        )

    return src_frame


@pytest.fixture
def empty_framelink() -> FramelinkPipeline:
    return FramelinkPipeline()


@pytest.fixture
def initial_framelink(src_frame) -> tuple[FramelinkPipeline, FramelinkModel[T]]:
    """
    :return: pipeline with a src frame already attached
    """
    pipeline: FramelinkPipeline = FramelinkPipeline()
    src_frame_model = pipeline.model()(src_frame)
    return pipeline, src_frame_model



from unittest.mock import MagicMock

import pandas as pd
import polars as pl
import pytest as pytest

from framelink.core import FramelinkPipeline
from framelink.storage.core import _NoStorage


def test_no_persistence(initial_framelink):
    pipeline, src_frame = initial_framelink

    assert isinstance(pipeline.settings.default_storage, _NoStorage)

    wrapped_store = MagicMock(wraps=pipeline.settings.default_storage)
    pipeline.settings.default_storage = wrapped_store

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pl.DataFrame:
        df_out = ctx.ref(src_frame).head()
        pl_df = pl.from_pandas(df_out)
        return pl_df

    @pipeline.model()
    def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        """
        Returns the bottom 5 records of the src_model
        """
        df_out = ctx.ref(src_frame).tail()
        return df_out

    _ = pipeline.build(tail_model)
    assert wrapped_store.call_count > 0


@pytest.mark.skip(reason="todo")
def test_in_memory_store():
    pass


@pytest.mark.skip(reason="todo")
def test_pickle_store():
    pass


@pytest.mark.skip(reason="todo")
def test_override_cacheing():
    pass


@pytest.mark.skip(reason="todo")
def test_model_run_to_nearest_cache():
    pass



import logging
from dataclasses import dataclass, field
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader
from pathlib import Path
from typing import Optional, TYPE_CHECKING

import typer
from rich import print

if TYPE_CHECKING:
    from framelink.core import FramelinkPipeline

_app = typer.Typer()


@dataclass
class CliContext:
    fl_pipelines: dict[str, "FramelinkPipeline"] = field(default_factory=dict)
    verbose: bool = False
    pipeline_name: Optional[str] = None


CLI_CONTEXT = CliContext()


@_app.callback()
def set_up(
    main_file: Path = typer.Argument(..., exists=True, readable=True, writable=False, file_okay=True, dir_okay=False),
    verbose: bool = typer.Option(False, "-v"),
    pipeline_name: Optional[str] = typer.Option(None),
):
    """
    The main setup script that's run before all commands.
    """
    main_file = main_file.resolve()
    print(f"Using pipelines defined from {main_file}")
    _import_pipelines_from_file(main_file)

    logging.basicConfig(level=logging.INFO)
    if verbose:
        print("Using verbose output")
        print(f"{CLI_CONTEXT=}")
        logging.basicConfig(level=logging.DEBUG)

    if len(CLI_CONTEXT.fl_pipelines) > 1 and not pipeline_name:
        raise ValueError("More than one pipeline to work with, please pass in a pipeline name.")

    CLI_CONTEXT.verbose = verbose
    CLI_CONTEXT.pipeline_name = pipeline_name


@_app.command()
def list_models(graph: bool = typer.Option(False, "-g")):
    """
    lists the models in the current framelink
    """

    print(f"listing models {graph=}")
    print(f"Model names: {list(CLI_CONTEXT.fl_pipelines.values())[0].model_names}")


def _import_pipelines_from_file(main_file: Path) -> None:
    """
    This function will, given a python file, import the code. If it contains a Framelink pipeline we will have the
    contex available for inspection and execution.
    :param main_file: the path to the python file to be imported.
    """
    if main_file.suffix != ".py":
        raise ImportError(f"{main_file} is not a .py file and is unsupported in this version of framelink")

    loader = SourceFileLoader("scripts", str(main_file))
    spec = spec_from_loader("scripts", loader)
    if spec:
        mymodule = module_from_spec(spec)
        loader.exec_module(mymodule)
    else:
        raise ImportError("Failed import :(")


@_app.command()
def build(model_name: str = typer.Argument(...)):
    """

    :param model_name:
    :return:
    """
    print(f"building {model_name}")


if __name__ == "__main__":
    _app()



import abc
from abc import ABC
from pathlib import Path
from typing import Any, Generic, Optional, TYPE_CHECKING

from framelink.types import T

if TYPE_CHECKING:
    from framelink.core import FramelinkModel, FramelinkPipeline


class FramelinkStorage(abc.ABC, Generic[T]):
    """
    The interface for interacting with a specific persistence medium for each model.
    """

    def __init__(self, lookup_from_store: bool = True) -> None:
        self.lookup_from_store = lookup_from_store

    @abc.abstractmethod
    def _store_frame(self, model: "FramelinkModel", result_frame: "T"):
        ...

    @abc.abstractmethod
    def _frame_lookup(self, model: "FramelinkModel", *args, **kwargs) -> Optional["T"]:
        ...

    def __call__(self, model: "FramelinkModel", ctx: "FramelinkPipeline", *args, **kwargs) -> "T":
        """
        This method should act as the retrieval function. A cache hit from `_store_frame` should return the stored
         frame, a cache miss will then just call the model and store its result using the `_store_frame` method.

        :param model: the model who may have a resulting frame to look up.
        :param ctx: the framelink context to execute against.
        :return: a cached frame or a new frame.
        """
        result = self._frame_lookup(model, *args, **kwargs) if self.lookup_from_store else None
        if result is None:
            result = model.callable(ctx)
            self._store_frame(model, result)
        return result


class FileStorage(FramelinkStorage, ABC):
    def __init__(self, data_dir: Path, file_suffix: str = ""):
        super().__init__()
        self.data_dir = data_dir
        self.file_suffix = file_suffix

    def _get_model_path(self, model: "FramelinkModel") -> Path:
        return self.data_dir / f"{model.name}.{self.file_suffix}"


class SQLPersistence(FramelinkStorage, ABC):
    def __init__(self, engine: Any):
        super().__init__()
        self.engine = engine



import pandas as pd

from framelink.core import FramelinkPipeline


def test_model_run(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    res = pipeline.build(head_model)
    assert len(res) == 5
    assert type(res) == pd.DataFrame
    assert head_model.call_count == 1



import random
import uuid

import networkx as nx
import pandas as pd
import pytest as pytest

from framelink.core import FramelinkModel, FramelinkPipeline


def test_model_link_dag(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def src_model_2(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        data = {
            "id_col": list(range(n)),
            "colour_col": [random.choice(["red", "green", "blue"]) for _ in range(n)],
            "uuid_col": [str(uuid.uuid4()) for _ in range(n)],
            "int_col": [random.randint(0, 100) for _ in range(n)],
        }
        return pd.DataFrame(data)

    src_model_2: FramelinkModel[pd.DataFrame]

    @pipeline.model()
    def only_blue_records(ctx: FramelinkPipeline) -> pd.DataFrame:
        src_2 = ctx.ref(src_model_2)
        src_2_blue = src_2.loc[src_2["colour_col"] == "blue", :]
        return src_2_blue

    @pipeline.model()
    def simple_src_1_filter(ctx: FramelinkPipeline) -> pd.DataFrame:
        frame = ctx.ref(src_frame)
        frame = frame.loc[:10]
        return frame

    @pipeline.model()
    def simple_src_1_filter2(ctx: FramelinkPipeline) -> pd.DataFrame:
        frame = ctx.ref(only_blue_records)
        ctx.ref(simple_src_1_filter)
        return frame

    assert nx.is_directed(pipeline.graph)
    assert nx.is_directed_acyclic_graph(pipeline.graph)
    assert src_model_2 in only_blue_records.upstreams
    assert simple_src_1_filter in src_frame.downstreams
    assert simple_src_1_filter2 in only_blue_records.downstreams
    assert simple_src_1_filter2 in simple_src_1_filter.downstreams


def test_deg_raises_error_on_premature_import(initial_framelink):
    pipeline, _ = initial_framelink

    with pytest.raises(KeyError) as e:

        @pipeline.model()
        def model_1(ctx: FramelinkPipeline):
            m_2 = ctx.ref(model_2)
            return m_2.head()

        @pipeline.model()
        def model_2(ctx: FramelinkPipeline):
            m_1 = ctx.ref(model_1)
            return m_1.head

    assert "model_2" in str(e.value)
    assert "could not locate" in str(e.value).lower()


@pytest.mark.skip(reason="todo")
def test_model_run_downstream():
    pass



import contextlib
import inspect
import logging
import textwrap
import time
from dataclasses import dataclass
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Callable, Collection, Generator, Generic, Iterator, Optional, Protocol, Union

import networkx as nx
import pydot

from framelink._cli import CLI_CONTEXT
from framelink._util import parse_model_src_for_internal_refs
from framelink.storage.core import NoStorage
from framelink.storage.interfaces import FramelinkStorage
from framelink.types import F, T


@dataclass
class FramelinkSettings:
    """Settings to be applied by the"""

    persist_models_dir: Path = Path(__file__).parent.parent / "data"
    default_log_level: int = logging.WARNING
    default_storage: FramelinkStorage = NoStorage()


class _FramelinkComponent(Protocol):
    _name: str

    @property
    def __name__(self):
        return self._name

    @property
    def __loc__(self):
        return hex(id(self))

    @property
    def name(self) -> str:
        """
        Util method for __name__ for natural flow in writing code.
        """
        return self.__name__

    def __key(self) -> tuple[Any, ...]:
        ...

    def __hash__(self) -> int:
        return hash(self.__key())


class FramelinkModel(_FramelinkComponent, Generic[T]):
    """
    A wrapper around a callable model that enables:
     - visibility onto the model DAG
     - caching of the model run
     - monitoring of run performance
     - adapters to run the model on various engines
    """

    callable: F[T]
    _graph_ref: nx.DiGraph
    call_perf: tuple[float, ...] = tuple()

    def __init__(
        self,
        model_func: F[T],
        graph: nx.DiGraph,
        pipeline_settings: FramelinkSettings,
        *,
        logging_level: Optional[int] = None,
        store: FramelinkStorage[T] = NoStorage(),
    ):
        # These are more "model settings"
        self.callable = model_func
        self._name = model_func.__name__
        self._store = store

        # These are the core attributes of the Model
        self._graph_ref = graph
        self._log = logging.getLogger(self.name)
        self._log.setLevel(logging_level if logging_level else pipeline_settings.default_log_level)

    def __repr__(self):
        return f"<{self.name} at {self.__loc__}>"

    @property
    def upstreams(self) -> set["FramelinkModel"]:
        """
        Pull this model from the graph and pull out all models are direct predecessors to this model.
        :return: a set of models that this model depends on.
        """
        return set(self._graph_ref.predecessors(self))

    @property
    def downstreams(self) -> set["FramelinkModel"]:
        """
        Pull this model from the graph and pull out all models are direct sucessors to this model.
        :return: a set of models that depend on this model.
        """
        return set(self._graph_ref.successors(self))

    @property
    def docstring(self) -> Optional[str]:
        """ """
        doc__ = self.callable.__doc__
        return textwrap.dedent(doc__).strip() if doc__ else None

    @property
    def source(self) -> str:
        """ """
        source__ = inspect.getsource(self.callable)
        source__ = textwrap.dedent(source__).strip()
        return source__

    @property
    def call_count(self) -> int:
        """return the numer of times this model has been called"""
        return len(self.call_perf)

    @property
    def perf_stats(self) -> tuple[float, ...]:
        """ """
        return self.call_perf

    @contextlib.contextmanager
    def _own_logging(self, ctx: "FramelinkPipeline"):
        old_log, ctx.log = ctx.log, self._log
        yield
        ctx.log = old_log

    def build(self, ctx: "FramelinkPipeline") -> T:
        """
        Build the current model with the context of the pipeline.
        :param ctx: "FramelinkPipeline": framelink pipeline context
        """
        with self._own_logging(ctx):
            ctx.log.info(f"Building {self.name}...")

            start_time = time.perf_counter()
            res = self._store(self, ctx)
            # question: this timer is the whole stream, how can we isolate just the build time for this model?
            build_time = time.perf_counter() - start_time
            ctx.log.info(f"Finished building {self.name} in {build_time:.3}s")

            # post build steps
            self.call_perf += (build_time,)
        return res

    def __key(self) -> tuple[str, Optional[str]]:
        """
        The uniqueness of a Model should be defined as its config and execution. This is used to determine the cache
        settings for the model when it is run
        """
        return self.name, self.source


class FramelinkPipeline(_FramelinkComponent):
    """The core class for building DAGs of models and producing links of the results.

    Each model linked to the pipeline will have context onto their upstream and downstream dependencies.
    """

    _models: dict[str, FramelinkModel]
    graph: nx.DiGraph
    log: logging.Logger  # Placeholder for model logger injection
    _log: logging.Logger  # Used for pipeline logging.

    def __init__(self, name: str = "default", settings: FramelinkSettings = FramelinkSettings()):
        super().__init__()
        self._name = name
        self._models = dict()
        self.graph = nx.DiGraph()
        self.settings = settings
        self._log = logging.getLogger(self.name)
        self._log.setLevel(settings.default_log_level)
        self.log = self._log
        CLI_CONTEXT.fl_pipelines[self._name] = self

    def __repr__(self):
        return f"<{self.name} with {len(self)} models at {self.__loc__}>"

    @property
    def model_names(self) -> list[str]:
        """Return a list of model names registered to this pipeline"""
        return sorted(m.name for m in self._models.values())

    @property
    def trees(self):
        return

    def graph_dot(self) -> pydot.Dot:
        """
        Using networkx and graphviz, create a DOT string representation of the model DAG.
        """
        return nx.drawing.nx_pydot.to_pydot(self.graph)

    def graph_plt(self):
        """
        Using networkx and matplotlib create an image representation of the model DAG.
        """
        # question: is there a better graph layout
        # pos = nx.multipartite_layout(self.graph)
        pos = nx.planar_layout(self.graph)
        return nx.draw_networkx(self.graph, pos)

    def model(
        self, *, logging_level=None, storage: Optional[FramelinkStorage[T]] = None
    ) -> Callable[[F[T]], FramelinkModel[T]]:
        """
        Annotation to register a model to the framelink pipeline.

        :param logging_level: Sets the logging level specifically for this model. If no level is passed it will default
            to the default level as per the pipelines settings.
        :param storage: Sets the persistence approach used when storing the model.
        """
        if storage:
            model_store_unwrapped = storage
        else:
            model_store_unwrapped = self.settings.default_storage

        def _decorator(func: F[T]) -> FramelinkModel[T]:
            """Internal wrapping of the model function to produce the metadata about the model.

            :param func: "PYPE_MODEL": the callable function that defines the model
            :returns: the wrapped model with all the extra pieces
            """

            model_wrapper: FramelinkModel = FramelinkModel(
                func, self.graph, self.settings, logging_level=logging_level, store=model_store_unwrapped
            )
            self._log.info(f"Registering model '{model_wrapper.name}'")

            # we need to keep a ref to the underlying graph to access the models when we ask for them via
            # `ref()` or `build()`
            assert model_wrapper.name not in self._models.keys(), f"Model already registered under {model_wrapper.name}"
            self._models[model_wrapper.name] = model_wrapper

            # todo: brainstorm more new ways of doing this.
            matches = parse_model_src_for_internal_refs(model_wrapper.source)
            matched_models = (self.get(name) for name in matches)

            self.graph.add_node(model_wrapper)
            upstream_edges = ((upstream_model, model_wrapper) for upstream_model in matched_models)
            self.graph.add_edges_from(upstream_edges)
            self._log.debug(
                f"Model '{model_wrapper.name}' has {len(model_wrapper.upstreams)} upstreams: {model_wrapper.upstreams}"
            )

            if not nx.is_directed_acyclic_graph(self.graph):
                cycle = nx.find_cycle(self.graph, model_wrapper)
                raise ValueError(f"{model_wrapper.name} has a loop: {cycle}")

            return model_wrapper

        return _decorator

    def ref(self, model: Union[FramelinkModel[T], str]) -> T:
        """
        ref will return the (cached) frame result of the model, so you can extend the frame inside another model.

        Raises a KeyError if the model Key cant be found.

        :param model: _Model: The model function with output you want to use.
        Example:
        >>> import pandas as pd
        >>>
        >>> pipeline = FramelinkPipeline()
        >>>
        >>> @pipeline.model()
        >>> def my_model_1(_: FramelinkPipeline) -> pd.DataFrame:
        >>>     return pd.read_csv("path/to/file.csv")
        >>>
        >>> @pipeline.model()
        >>> def my_model_2(ctx: FramelinkPipeline) -> pd.DataFrame:
        >>>     return ctx.ref(my_model_1).head()
        """
        model_wrapper = self.get(model)
        return model_wrapper.build(self)

    def build(self, model_name: Union[FramelinkModel[T], str]) -> T:
        """Building models is just proxied through to ref. Each build command should build only the given node in the
         graph up to the nearest cache or persisted cache.

        :param model_name: "PYPE_MODEL": the model to build in the context of this pipeline.
        :param overrides: A mapping of models whos result should be overridden for this build.
        """
        return self.ref(model_name)

    def topological_sorted_nodes(self) -> Generator[tuple[FramelinkModel, ...], None, None]:
        """
        This implementation is more parallel aware than the networkx implementation/

        Basically a repeatable version of the example at https://docs.python.org/3/library/graphlib.html
        :return:
        """
        topological_sorter: TopologicalSorter = TopologicalSorter()
        for node in self._models.values():
            topological_sorter.add(node, *node.upstreams)

        topological_sorter.prepare()
        while topological_sorter.is_active():
            nodes = topological_sorter.get_ready()
            yield nodes
            topological_sorter.done(*nodes)

    @classmethod
    def merge_pipelines(
        cls,
        new_name: str,
        settings: FramelinkSettings = FramelinkSettings(),
        *,
        collection: Collection["FramelinkPipeline"],
    ) -> "FramelinkPipeline":
        new_pipeline = FramelinkPipeline(new_name, settings)

        for other in collection:
            print(other.name)
            assert len(set(new_pipeline.model_names) & set(other.model_names)) == 0
            new_pipeline._models = new_pipeline._models | other._models

        return new_pipeline

    def __key(self) -> tuple[int, ...]:
        """
        The state of a pypeline should be the aggregation settings that determine the way models should behave.

        When we cache a model frame we need to hash the context for its run, which links to settings it should
        take while building.
        """
        return tuple(hash(m) for m in self._models.values())

    def __len__(self) -> int:
        return len(self._models.keys())

    def __iter__(self) -> Iterator[FramelinkModel]:
        return self._models.values().__iter__()

    def __contains__(self, item: Union[FramelinkModel, F[T]]) -> bool:
        return item.__name__ in self._models.keys() or item in self._models.values()

    def get(self, model: Union[FramelinkModel, str]) -> FramelinkModel:
        """
        Given a `FramelinkModel`, or the model's name, return the`FramelinkModel`.


        :param model: Model function or the model name (function name).
        :returns: The `FramelinkModel` that wraps the model.
        """

        try:
            model_name = model if type(model) == str else model.name  # type: ignore
            return self._models[model_name]
        except (KeyError, AttributeError) as k:
            raise KeyError(
                f"Could not locate the model '{model}' in the pipeline {self.name} models: {self.model_names}. "
                "Have you registered it correctly?"
            ) from k



import uuid
from typing import Generator

import networkx as nx
import pandas as pd
import pytest

from framelink.core import FramelinkPipeline


def test_merge_pipelines_missing_merge(src_frame):
    pipeline_1 = FramelinkPipeline("pipeline_1")
    pipeline_2 = FramelinkPipeline("pipeline_2")

    src_frame = pipeline_1.model()(src_frame)

    @pipeline_1.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    with pytest.raises(KeyError) as ke:

        @pipeline_2.model()
        def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:
            df_out = ctx.ref(src_frame).tail()
            return df_out

    assert "src_frame" in str(ke.value)
    assert "could not locate" in str(ke.value).lower()


def test_topological_sorted_nodes():
    pipeline = FramelinkPipeline()

    @pipeline.model()
    def src_frame_1(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        return pd.DataFrame({"id_col": [uuid.uuid4() for _ in range(n)]})

    @pipeline.model()
    def src_frame_2(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        return pd.DataFrame({"id_col2": [uuid.uuid4() for _ in range(n)]})

    @pipeline.model()
    def merge_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        src_1 = ctx.ref(src_frame_1)
        src_2 = ctx.ref(src_frame_2)
        return pd.merge(src_1, src_2)

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        merged = ctx.ref(merge_model)
        return merged.head()

    assert len(pipeline) == 4

    ts = pipeline.topological_sorted_nodes()

    assert isinstance(ts, Generator)

    ts_l = list(ts)
    assert len(ts_l) == 3
    assert all(type(e) == tuple for e in ts_l)
    assert ts_l == [(src_frame_1, src_frame_2), (merge_model,), (head_model,)]


@pytest.mark.xfail(reason="joining isn't finished yet")
def test_merge_pipelines_distinct():
    pipeline_1 = FramelinkPipeline("pipeline_1")
    pipeline_2 = FramelinkPipeline("pipeline_2")

    @pipeline_1.model()
    def src_frame_1(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        return pd.DataFrame({"id_col": [uuid.uuid4() for _ in range(n)]})

    @pipeline_1.model()
    def src_frame_2(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        return pd.DataFrame({"id_col2": [uuid.uuid4() for _ in range(n)]})

    @pipeline_1.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        src_1 = ctx.ref(src_frame_1)
        src_2 = ctx.ref(src_frame_2)
        merged = pd.merge(src_1, src_2)
        return merged.head()

    @pipeline_2.model()
    def src_frame_3(_: FramelinkPipeline) -> pd.DataFrame:
        n = 10
        return pd.DataFrame({"id_col3": [uuid.uuid4() for _ in range(n)]})

    @pipeline_2.model()
    def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame_3).tail()
        return df_out

    res = FramelinkPipeline.merge_pipelines("merged", collection=[pipeline_1, pipeline_2])

    ts_1_g = pipeline_1.topological_sorted_nodes()
    list(ts_1_g)
    list(pipeline_2.topological_sorted_nodes())

    assert res.name == "merged"
    assert nx.is_directed_acyclic_graph(res.graph)
    assert nx.number_weakly_connected_components(pipeline_1.graph) == 1
    assert nx.number_weakly_connected_components(pipeline_2.graph) == 1
    assert nx.number_weakly_connected_components(res.graph) == 2
    assert len(res) == 4
    assert all(
        n in res.model_names
        for n in (
            "src_frame_1",
            "head_model",
            "src_frame_2",
            "tail_model",
        )
    )
    assert head_model in src_frame_1.downstreams
    assert src_frame_1 in head_model.upstreams
    assert tail_model in src_frame_2.downstreams
    assert src_frame_2 in tail_model.upstreams



import pandas as pd
import polars as pl
import pytest

from framelink.core import FramelinkModel, FramelinkPipeline


def test_model_registration(empty_framelink):
    pipeline = empty_framelink

    @pipeline.model()
    def src_frame(_: FramelinkPipeline) -> pl.LazyFrame:
        """
        Im a docstring!
        """
        return pl.LazyFrame()

    # naming things
    assert src_frame.name == "src_frame"
    assert src_frame.name == src_frame.__name__

    # metadata of model
    assert "Im a docstring!" in src_frame.docstring
    assert src_frame.call_count == 0
    assert src_frame.perf_stats == tuple()

    # model settings
    assert src_frame._log is not None

    # association to pipeline
    assert src_frame in pipeline
    assert src_frame.name in pipeline.model_names


def test_model_linking_linear(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def model_1(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    @pipeline.model()
    def model_2(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(model_1).sort_values("int_col").tail(2)
        return df_out

    assert len(pipeline) == 3  # including the initial one
    assert model_1 in pipeline
    assert model_2 in pipeline
    assert pipeline.model_names == ["model_1", "model_2", "src_frame"]
    assert src_frame in model_1.upstreams
    assert model_2 in model_1.downstreams


def test_model_ref(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    def model_tail(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).tail()
        return df_out

    # basic assertions for registered models
    assert len(pipeline) == 2
    assert type(head_model) == FramelinkModel
    assert type(src_frame) == FramelinkModel
    assert head_model in pipeline
    assert src_frame in pipeline
    assert pipeline.model_names == ["head_model", "src_frame"]

    # ref models
    ref_res = pipeline.ref(head_model)
    assert ref_res is not None
    assert len(ref_res) == 5

    # missing models
    assert model_tail not in pipeline
    with pytest.raises(KeyError) as e1:
        pipeline.ref(model_tail)

    assert "model_tail" in str(e1.value)

    # missing intermediate model.
    with pytest.raises(KeyError) as e2:

        @pipeline.model()
        def downstream_model(ctx: FramelinkPipeline):
            return ctx.ref(model_tail).head(2)

        pipeline.build(downstream_model)

    assert "model_tail" in str(e2.value)


def test_get_metadata(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    @pipeline.model()
    def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        """
        Returns the bottom 5 records of the src_model
        """
        df_out = ctx.ref(src_frame).tail()
        return df_out

    assert head_model in pipeline
    assert tail_model in pipeline
    assert head_model.docstring is None
    assert tail_model.docstring == "Returns the bottom 5 records of the src_model"
    assert tail_model.source == (
        "@pipeline.model()\n"
        "def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:\n"
        '    """\n'
        "    Returns the bottom 5 records of the src_model\n"
        '    """\n'
        "    df_out = ctx.ref(src_frame).tail()\n"
        "    return df_out"
    )


def test_type_values(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    assert isinstance(head_model, FramelinkModel)

    def head_model_2(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    assert not isinstance(head_model_2, FramelinkModel)
    head_model_2_wrapped = pipeline.model()(head_model_2)
    assert isinstance(head_model_2_wrapped, FramelinkModel)


def test_get_lookups(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref(src_frame).head()
        return df_out

    lookup_model = pipeline.get(head_model)
    assert lookup_model
    assert lookup_model is head_model

    lookup_str = pipeline.get("head_model")
    assert lookup_str
    assert lookup_str is head_model


def test_ref_string(initial_framelink):
    pipeline, src_frame = initial_framelink

    @pipeline.model()
    def head_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref("src_frame").head()
        return df_out

    # black should keep the quotes in the below model as '
    # fmt: off
    @pipeline.model()
    def tail_model(ctx: FramelinkPipeline) -> pd.DataFrame:
        df_out = ctx.ref('src_frame').head()
        return df_out

    # fmt: on

    assert head_model
    assert head_model in src_frame.downstreams
    assert src_frame in head_model.upstreams



import re


def parse_model_src_for_internal_refs(model_src: str) -> set[str]:
    """
    This utility is for extracting the names any time a model has the line `.ref(<model_name>)` in. It's a util not part
        of the FramelinkModel as we don't really want this to be called more than once when registering the model. Once
        a model is registered we can identify the `.ref`d models with `.upstreams`.

    :param model_src: the FramelinkModel model source to search through.
    :return: a list of all the found model names which are `ref`d in the model source.
    """
    pattern = r"\.ref\(['\"]?(.*?)['\"]?\)"
    matches = re.findall(pattern, model_src)
    return set(matches)



import functools
import pickle
from pathlib import Path
from typing import Optional, TYPE_CHECKING

from framelink.storage.interfaces import FileStorage, FramelinkStorage
from framelink.types import T

if TYPE_CHECKING:
    from framelink.core import FramelinkModel


class _NoStorage(FramelinkStorage):
    def __init__(self) -> None:
        super().__init__()

    def _store_frame(self, model: "FramelinkModel", result_frame: "T"):
        pass

    def _frame_lookup(self, model: "FramelinkModel", *args, **kwargs) -> None:
        return None


_no_storage_singleton = _NoStorage()


def NoStorage() -> _NoStorage:
    return _no_storage_singleton


class InMemory(FramelinkStorage):
    def __init__(self, lookup_from_store: bool = True) -> None:
        super().__init__(lookup_from_store)
        self._cache = functools.lru_cache()

    def _store_frame(self, model: "FramelinkModel", result_frame: "T"):
        # This is stored for us by the lru cache
        pass

    def _frame_lookup(self, model: "FramelinkModel", *args, **kwargs) -> Optional["T"]:
        cache = self._cache(model.build)
        return cache()


class PickleStorage(FileStorage):
    def __init__(self, data_dir: Path):
        super().__init__(data_dir, "pickle")

    def _store_frame(self, model: "FramelinkModel", result_frame: "T"):
        path = self._get_model_path(model)
        with path.open("wb") as f:
            pickle.dumps(f)

    def _frame_lookup(self, model: "FramelinkModel", *args, **kwargs) -> Optional["T"]:
        path = self._get_model_path(model)
        with path.open("rb") as f:
            res = pickle.load(f)
        return res



from typing import Callable, TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from framelink.core import FramelinkPipeline

T = TypeVar("T")  # usually one of pl.DataFrame, pl.DataFrame, pl.LazyFrame.. etc
F = Callable[["FramelinkPipeline"], T]  # the definition of one of the models users will write



import pytest

from framelink._util import parse_model_src_for_internal_refs


@pytest.mark.parametrize(
    "in_str, model_names",
    [
        (".ref(%s)", ("my_model",)),
        (".ref('%s')", ("my_model",)),
        ('.ref("%s")', ("my_model",)),
        ('ref.ref("%s")', ("my_model",)),
        ('.ref("%s").ref(%s)', ("my_model", "model_two")),
        ('.ref("%s").ref(%s)', ("my_ref_model", "ref_model")),
    ],
)
def test_model_body_ref_parsing(in_str, model_names):
    in_str = in_str % model_names  # format the string with the given model name
    res = parse_model_src_for_internal_refs(in_str)
    assert len(res) == len(model_names)
    assert all(model_name in res for model_name in model_names)



import logging

import pandas as pd

from framelink.core import FramelinkPipeline, FramelinkSettings


def test_basic_logging(src_frame, caplog):
    logging.basicConfig(level=logging.DEBUG)
    model_1_msg = "message from model_1!"
    model_1_msg_2 = "second part!"
    model_2_msg = "message from model_2!"

    pipeline = FramelinkPipeline(settings=FramelinkSettings(default_log_level=logging.INFO))
    src_frame = pipeline.model()(src_frame)
    og_logger = pipeline.log

    @pipeline.model()  # will be INFO level, as per default passed
    def model_1(ctx: FramelinkPipeline) -> pd.DataFrame:
        model_1_log = ctx.log
        assert ctx.log != og_logger

        ctx.log.warning(model_1_msg)
        df_out = ctx.ref(src_frame).head()

        assert ctx.log == model_1_log
        ctx.log.debug(model_1_msg_2)

        return df_out

    @pipeline.model(logging_level=logging.DEBUG)
    def model_2(ctx: FramelinkPipeline) -> pd.DataFrame:
        ctx.log.info(model_2_msg + " info!")
        ctx.log.debug(model_2_msg + " debug!")

        assert ctx.log != og_logger
        model_2_log = ctx.log

        df_out = ctx.ref(model_1).sort_values("int_col").tail(2)

        assert ctx.log == model_2_log
        return df_out

    all_messages = ", ".join(caplog.messages).lower()
    assert "register" in all_messages and "model_1" in all_messages
    assert "register" in all_messages and "model_2" in all_messages
    assert model_2_msg not in all_messages

    assert pipeline.log == og_logger
    pipeline.build(model_2)
    assert pipeline.log == og_logger
    all_messages_post_build = ", ".join(caplog.messages).lower()

    print(caplog.record_tuples)
    assert model_2_msg + " debug!" in all_messages_post_build
